@@ -1,4 +1,3 @@
-use std::cell::RefCell;
 use std::collections::BTreeMap;
 use std::str;
 use std::sync::atomic::{AtomicUsize, Ordering};
@@ -13,62 +12,65 @@ use syscall::flag::{
     EventFlags, MODE_PERM, O_ACCMODE, O_CREAT, O_DIRECTORY, O_EXCL, O_NOFOLLOW, O_RDONLY, O_RDWR,
     O_STAT, O_SYMLINK, O_TRUNC, O_WRONLY,
 };
-use syscall::scheme::Scheme;
+use syscall::scheme::SchemeMut;
 
-use disk::Disk;
-use filesystem::FileSystem;
-use node::Node;
-use BLOCK_SIZE;
+use crate::{Disk, FileSystem, Node, Transaction, TreeData, TreePtr, BLOCK_SIZE};
 
 use super::resource::{DirResource, FileResource, Resource};
 
 pub struct FileScheme<D: Disk> {
     name: String,
-    fs: RefCell<FileSystem<D>>,
+    fs: FileSystem<D>,
     next_id: AtomicUsize,
-    files: RefCell<BTreeMap<usize, Box<Resource<D>>>>,
-    fmap: RefCell<BTreeMap<usize, usize>>,
+    files: BTreeMap<usize, Box<dyn Resource<D>>>,
+    fmap: BTreeMap<usize, usize>,
 }
 
 impl<D: Disk> FileScheme<D> {
     pub fn new(name: String, fs: FileSystem<D>) -> FileScheme<D> {
         FileScheme {
             name: name,
-            fs: RefCell::new(fs),
+            fs: fs,
             next_id: AtomicUsize::new(1),
-            files: RefCell::new(BTreeMap::new()),
-            fmap: RefCell::new(BTreeMap::new()),
+            files: BTreeMap::new(),
+            fmap: BTreeMap::new(),
         }
     }
 
     fn resolve_symlink(
-        &self,
-        fs: &mut FileSystem<D>,
+        scheme_name: &str,
+        tx: &mut Transaction<D>,
         uid: u32,
         gid: u32,
-        url: &[u8],
-        node: (u64, Node),
-        nodes: &mut Vec<(u64, Node)>,
+        url: &str,
+        node: TreeData<Node>,
+        nodes: &mut Vec<(TreeData<Node>, String)>,
     ) -> Result<Vec<u8>> {
+        let atime = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
+
         let mut node = node;
         for _ in 0..32 {
             // XXX What should the limit be?
-            let atime = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
             let mut buf = [0; 4096];
-            let count = fs.read_node(node.0, 0, &mut buf, atime.as_secs(), atime.subsec_nanos())?;
-            let scheme = format!("{}:", &self.name);
-            let canon = canonicalize(
-                &format!("{}{}", scheme, str::from_utf8(url).unwrap()).as_bytes(),
-                &buf[0..count],
-            );
+            let count = tx.read_node(
+                node.ptr(),
+                0,
+                &mut buf,
+                atime.as_secs(),
+                atime.subsec_nanos(),
+            )?;
+            let scheme = format!("{}:", scheme_name);
+            let canon = canonicalize(&format!("{}{}", scheme, url).as_bytes(), &buf[0..count]);
             let path = str::from_utf8(&canon[scheme.len()..])
                 .unwrap_or("")
                 .trim_matches('/');
             nodes.clear();
-            if let Some(next_node) = self.path_nodes(fs, path, uid, gid, nodes)? {
-                if !next_node.1.is_symlink() {
+            if let Some((next_node, next_node_name)) =
+                Self::path_nodes(scheme_name, tx, path, uid, gid, nodes)?
+            {
+                if !next_node.data().is_symlink() {
                     if canon.starts_with(scheme.as_bytes()) {
-                        nodes.push(next_node);
+                        nodes.push((next_node, next_node_name));
                         return Ok(canon[scheme.len()..].to_vec());
                     } else {
                         return Err(Error::new(EXDEV));
@@ -83,47 +85,51 @@ impl<D: Disk> FileScheme<D> {
     }
 
     fn path_nodes(
-        &self,
-        fs: &mut FileSystem<D>,
+        scheme_name: &str,
+        tx: &mut Transaction<D>,
         path: &str,
         uid: u32,
         gid: u32,
-        nodes: &mut Vec<(u64, Node)>,
-    ) -> Result<Option<(u64, Node)>> {
+        nodes: &mut Vec<(TreeData<Node>, String)>,
+    ) -> Result<Option<(TreeData<Node>, String)>> {
         let mut parts = path.split('/').filter(|part| !part.is_empty());
-        let mut part_opt = None;
-        let mut block = fs.header.1.root;
+        let mut part_opt: Option<&str> = None;
+        let mut node_ptr = TreePtr::root();
+        let mut node_name = String::new();
         loop {
             let node_res = match part_opt {
-                None => fs.node(block),
-                Some(part) => fs.find_node(part, block),
+                None => tx.read_tree(node_ptr),
+                Some(part) => {
+                    node_name = part.to_string();
+                    tx.find_node(node_ptr, part)
+                }
             };
 
             part_opt = parts.next();
-            if part_opt.is_some() {
+            if let Some(part) = part_opt {
                 let node = node_res?;
-                if !node.1.permission(uid, gid, Node::MODE_EXEC) {
+                if !node.data().permission(uid, gid, Node::MODE_EXEC) {
                     return Err(Error::new(EACCES));
                 }
-                if node.1.is_symlink() {
-                    let mut url = Vec::new();
-                    url.extend_from_slice(self.name.as_bytes());
-                    url.push(b':');
-                    for i in nodes.iter() {
-                        url.push(b'/');
-                        url.extend_from_slice(&i.1.name);
+                if node.data().is_symlink() {
+                    let mut url = String::new();
+                    url.push_str(scheme_name);
+                    url.push(':');
+                    for (_parent, parent_name) in nodes.iter() {
+                        url.push('/');
+                        url.push_str(&parent_name);
                     }
-                    self.resolve_symlink(fs, uid, gid, &url, node, nodes)?;
-                    block = nodes.last().unwrap().0;
-                } else if !node.1.is_dir() {
+                    Self::resolve_symlink(scheme_name, tx, uid, gid, &url, node, nodes)?;
+                    node_ptr = nodes.last().unwrap().0.ptr();
+                } else if !node.data().is_dir() {
                     return Err(Error::new(ENOTDIR));
                 } else {
-                    block = node.0;
-                    nodes.push(node);
+                    node_ptr = node.ptr();
+                    nodes.push((node, part.to_string()));
                 }
             } else {
                 match node_res {
-                    Ok(node) => return Ok(Some(node)),
+                    Ok(node) => return Ok(Some((node, node_name))),
                     Err(err) => match err.errno {
                         ENOENT => return Ok(None),
                         _ => return Err(err),
@@ -204,96 +210,110 @@ pub fn canonicalize(current: &[u8], path: &[u8]) -> Vec<u8> {
     }
 }
 
-impl<D: Disk> Scheme for FileScheme<D> {
-    fn open(&self, url: &str, flags: usize, uid: u32, gid: u32) -> Result<usize> {
+impl<D: Disk> SchemeMut for FileScheme<D> {
+    fn open(&mut self, url: &str, flags: usize, uid: u32, gid: u32) -> Result<usize> {
         let path = url.trim_matches('/');
 
         // println!("Open '{}' {:X}", path, flags);
 
-        let mut fs = self.fs.borrow_mut();
-
+        //TODO: try to move things into one transaction
+        let scheme_name = &self.name;
         let mut nodes = Vec::new();
-        let node_opt = self.path_nodes(&mut fs, path, uid, gid, &mut nodes)?;
-        let resource: Box<Resource<D>> = match node_opt {
-            Some(node) => {
+        let node_opt = self
+            .fs
+            .tx(|tx| Self::path_nodes(scheme_name, tx, path, uid, gid, &mut nodes))?;
+        let resource: Box<dyn Resource<D>> = match node_opt {
+            Some((node, _node_name)) => {
                 if flags & (O_CREAT | O_EXCL) == O_CREAT | O_EXCL {
                     return Err(Error::new(EEXIST));
-                } else if node.1.is_dir() {
+                } else if node.data().is_dir() {
                     if flags & O_ACCMODE == O_RDONLY {
-                        if !node.1.permission(uid, gid, Node::MODE_READ) {
-                            // println!("dir not readable {:o}", node.1.mode);
+                        if !node.data().permission(uid, gid, Node::MODE_READ) {
+                            // println!("dir not readable {:o}", node.data().mode);
                             return Err(Error::new(EACCES));
                         }
 
                         let mut children = Vec::new();
-                        fs.child_nodes(&mut children, node.0)?;
+                        self.fs.tx(|tx| tx.child_nodes(node.ptr(), &mut children))?;
 
                         let mut data = Vec::new();
                         for child in children.iter() {
-                            if let Ok(name) = child.1.name() {
+                            if let Some(child_name) = child.name() {
                                 if !data.is_empty() {
                                     data.push(b'\n');
                                 }
-                                data.extend_from_slice(&name.as_bytes());
+                                data.extend_from_slice(&child_name.as_bytes());
                             }
                         }
 
-                        Box::new(DirResource::new(path.to_string(), node.0, Some(data), uid))
+                        Box::new(DirResource::new(
+                            path.to_string(),
+                            node.ptr(),
+                            Some(data),
+                            uid,
+                        ))
                     } else if flags & O_WRONLY == O_WRONLY {
                         // println!("{:X} & {:X}: EISDIR {}", flags, O_DIRECTORY, path);
                         return Err(Error::new(EISDIR));
                     } else {
-                        Box::new(DirResource::new(path.to_string(), node.0, None, uid))
+                        Box::new(DirResource::new(path.to_string(), node.ptr(), None, uid))
                     }
-                } else if node.1.is_symlink()
+                } else if node.data().is_symlink()
                     && !(flags & O_STAT == O_STAT && flags & O_NOFOLLOW == O_NOFOLLOW)
                     && flags & O_SYMLINK != O_SYMLINK
                 {
                     let mut resolve_nodes = Vec::new();
-                    let resolved = self.resolve_symlink(
-                        &mut fs,
-                        uid,
-                        gid,
-                        url.as_bytes(),
-                        node,
-                        &mut resolve_nodes,
-                    )?;
-                    drop(fs);
+                    let resolved = self.fs.tx(|tx| {
+                        Self::resolve_symlink(
+                            scheme_name,
+                            tx,
+                            uid,
+                            gid,
+                            url,
+                            node,
+                            &mut resolve_nodes,
+                        )
+                    })?;
                     let resolved_utf8 =
                         str::from_utf8(&resolved).map_err(|_| Error::new(EINVAL))?;
                     return self.open(resolved_utf8, flags, uid, gid);
-                } else if !node.1.is_symlink() && flags & O_SYMLINK == O_SYMLINK {
+                } else if !node.data().is_symlink() && flags & O_SYMLINK == O_SYMLINK {
                     return Err(Error::new(EINVAL));
                 } else {
+                    let node_ptr = node.ptr();
+
                     if flags & O_DIRECTORY == O_DIRECTORY {
                         // println!("{:X} & {:X}: ENOTDIR {}", flags, O_DIRECTORY, path);
                         return Err(Error::new(ENOTDIR));
                     }
 
                     if (flags & O_ACCMODE == O_RDONLY || flags & O_ACCMODE == O_RDWR)
-                        && !node.1.permission(uid, gid, Node::MODE_READ)
+                        && !node.data().permission(uid, gid, Node::MODE_READ)
                     {
-                        // println!("file not readable {:o}", node.1.mode);
+                        // println!("file not readable {:o}", node.data().mode);
                         return Err(Error::new(EACCES));
                     }
 
                     if (flags & O_ACCMODE == O_WRONLY || flags & O_ACCMODE == O_RDWR)
-                        && !node.1.permission(uid, gid, Node::MODE_WRITE)
+                        && !node.data().permission(uid, gid, Node::MODE_WRITE)
                     {
-                        // println!("file not writable {:o}", node.1.mode);
+                        // println!("file not writable {:o}", node.data().mode);
                         return Err(Error::new(EACCES));
                     }
 
                     if flags & O_TRUNC == O_TRUNC {
-                        if !node.1.permission(uid, gid, Node::MODE_WRITE) {
-                            // println!("file not writable {:o}", node.1.mode);
+                        if !node.data().permission(uid, gid, Node::MODE_WRITE) {
+                            // println!("file not writable {:o}", node.data().mode);
                             return Err(Error::new(EACCES));
                         }
 
-                        fs.node_set_len(node.0, 0)?;
+                        let mtime = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
+                        self.fs.tx(|tx| {
+                            tx.truncate_node(node_ptr, 0, mtime.as_secs(), mtime.subsec_nanos())
+                        })?;
                     }
 
-                    Box::new(FileResource::new(path.to_string(), node.0, flags, uid))
+                    Box::new(FileResource::new(path.to_string(), node_ptr, flags, uid))
                 }
             }
             None => {
@@ -305,8 +325,8 @@ impl<D: Disk> Scheme for FileScheme<D> {
                         }
                     }
                     if !last_part.is_empty() {
-                        if let Some(parent) = nodes.last() {
-                            if !parent.1.permission(uid, gid, Node::MODE_WRITE) {
+                        if let Some((parent, _parent_name)) = nodes.last() {
+                            if !parent.data().permission(uid, gid, Node::MODE_WRITE) {
                                 // println!("dir not writable {:o}", parent.1.mode);
                                 return Err(Error::new(EACCES));
                             }
@@ -320,22 +340,28 @@ impl<D: Disk> Scheme for FileScheme<D> {
                                 Node::MODE_FILE
                             };
 
-                            let ctime = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
-                            let mut node = fs.create_node(
-                                mode_type | (flags as u16 & Node::MODE_PERM),
-                                &last_part,
-                                parent.0,
-                                ctime.as_secs(),
-                                ctime.subsec_nanos(),
-                            )?;
-                            node.1.uid = uid;
-                            node.1.gid = gid;
-                            fs.write_at(node.0, &node.1)?;
+                            let node_ptr = self.fs.tx(|tx| {
+                                let ctime = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
+                                let mut node = tx.create_node(
+                                    parent.ptr(),
+                                    &last_part,
+                                    mode_type | (flags as u16 & Node::MODE_PERM),
+                                    ctime.as_secs(),
+                                    ctime.subsec_nanos(),
+                                )?;
+                                let node_ptr = node.ptr();
+                                if node.data().uid() != uid || node.data().gid() != gid {
+                                    node.data_mut().set_uid(uid);
+                                    node.data_mut().set_gid(gid);
+                                    tx.sync_tree(node)?;
+                                }
+                                Ok(node_ptr)
+                            })?;
 
                             if dir {
-                                Box::new(DirResource::new(path.to_string(), node.0, None, uid))
+                                Box::new(DirResource::new(path.to_string(), node_ptr, None, uid))
                             } else {
-                                Box::new(FileResource::new(path.to_string(), node.0, flags, uid))
+                                Box::new(FileResource::new(path.to_string(), node_ptr, flags, uid))
                             }
                         } else {
                             return Err(Error::new(EPERM));
@@ -350,196 +376,194 @@ impl<D: Disk> Scheme for FileScheme<D> {
         };
 
         let id = self.next_id.fetch_add(1, Ordering::SeqCst);
-        self.files.borrow_mut().insert(id, resource);
+        self.files.insert(id, resource);
 
         Ok(id)
     }
 
-    fn chmod(&self, url: &str, mode: u16, uid: u32, gid: u32) -> Result<usize> {
+    fn chmod(&mut self, url: &str, mode: u16, uid: u32, gid: u32) -> Result<usize> {
         let path = url.trim_matches('/');
 
         // println!("Chmod '{}'", path);
 
-        let mut fs = self.fs.borrow_mut();
+        let scheme_name = &self.name;
+        self.fs.tx(|tx| {
+            let mut nodes = Vec::new();
+            if let Some((mut node, _node_name)) =
+                Self::path_nodes(scheme_name, tx, path, uid, gid, &mut nodes)?
+            {
+                if node.data().uid() == uid || uid == 0 {
+                    let old_mode = node.data().mode();
+                    let new_mode = (old_mode & !MODE_PERM) | (mode & MODE_PERM);
+                    if old_mode != new_mode {
+                        node.data_mut().set_mode(new_mode);
+                        tx.sync_tree(node)?;
+                    }
 
-        let mut nodes = Vec::new();
-        if let Some(mut node) = self.path_nodes(&mut fs, path, uid, gid, &mut nodes)? {
-            if node.1.uid == uid || uid == 0 {
-                node.1.mode = (node.1.mode & !MODE_PERM) | (mode & MODE_PERM);
-                fs.write_at(node.0, &node.1)?;
-                Ok(0)
+                    Ok(0)
+                } else {
+                    Err(Error::new(EPERM))
+                }
             } else {
-                Err(Error::new(EPERM))
+                Err(Error::new(ENOENT))
             }
-        } else {
-            Err(Error::new(ENOENT))
-        }
+        })
     }
 
-    fn rmdir(&self, url: &str, uid: u32, gid: u32) -> Result<usize> {
+    fn rmdir(&mut self, url: &str, uid: u32, gid: u32) -> Result<usize> {
         let path = url.trim_matches('/');
 
         // println!("Rmdir '{}'", path);
 
-        let mut fs = self.fs.borrow_mut();
-
-        let mut nodes = Vec::new();
-        if let Some(child) = self.path_nodes(&mut fs, path, uid, gid, &mut nodes)? {
-            if let Some(parent) = nodes.last() {
-                if !parent.1.permission(uid, gid, Node::MODE_WRITE) {
-                    // println!("dir not writable {:o}", parent.1.mode);
-                    return Err(Error::new(EACCES));
-                }
-
-                if child.1.is_dir() {
-                    if !child.1.permission(uid, gid, Node::MODE_WRITE) {
+        let scheme_name = &self.name;
+        self.fs.tx(|tx| {
+            let mut nodes = Vec::new();
+            if let Some((child, child_name)) =
+                Self::path_nodes(scheme_name, tx, path, uid, gid, &mut nodes)?
+            {
+                if let Some((parent, _parent_name)) = nodes.last() {
+                    if !parent.data().permission(uid, gid, Node::MODE_WRITE) {
                         // println!("dir not writable {:o}", parent.1.mode);
                         return Err(Error::new(EACCES));
                     }
 
-                    if let Ok(child_name) = child.1.name() {
-                        fs.remove_node(Node::MODE_DIR, child_name, parent.0)
+                    if child.data().is_dir() {
+                        if !child.data().permission(uid, gid, Node::MODE_WRITE) {
+                            // println!("dir not writable {:o}", parent.1.mode);
+                            return Err(Error::new(EACCES));
+                        }
+
+                        tx.remove_node(parent.ptr(), &child_name, Node::MODE_DIR)
                             .and(Ok(0))
                     } else {
-                        Err(Error::new(ENOENT))
+                        Err(Error::new(ENOTDIR))
                     }
                 } else {
-                    Err(Error::new(ENOTDIR))
+                    Err(Error::new(EPERM))
                 }
             } else {
-                Err(Error::new(EPERM))
+                Err(Error::new(ENOENT))
             }
-        } else {
-            Err(Error::new(ENOENT))
-        }
+        })
     }
 
-    fn unlink(&self, url: &str, uid: u32, gid: u32) -> Result<usize> {
+    fn unlink(&mut self, url: &str, uid: u32, gid: u32) -> Result<usize> {
         let path = url.trim_matches('/');
 
         // println!("Unlink '{}'", path);
 
-        let mut fs = self.fs.borrow_mut();
-
-        let mut nodes = Vec::new();
-        if let Some(child) = self.path_nodes(&mut fs, path, uid, gid, &mut nodes)? {
-            if let Some(parent) = nodes.last() {
-                if !parent.1.permission(uid, gid, Node::MODE_WRITE) {
-                    // println!("dir not writable {:o}", parent.1.mode);
-                    return Err(Error::new(EACCES));
-                }
-
-                if !child.1.is_dir() {
-                    if child.1.uid != uid {
-                        // println!("file not owned by current user {}", parent.1.uid);
+        let scheme_name = &self.name;
+        self.fs.tx(|tx| {
+            let mut nodes = Vec::new();
+            if let Some((child, child_name)) =
+                Self::path_nodes(scheme_name, tx, path, uid, gid, &mut nodes)?
+            {
+                if let Some((parent, _parent_name)) = nodes.last() {
+                    if !parent.data().permission(uid, gid, Node::MODE_WRITE) {
+                        // println!("dir not writable {:o}", parent.1.mode);
                         return Err(Error::new(EACCES));
                     }
 
-                    if let Ok(child_name) = child.1.name() {
-                        if child.1.is_symlink() {
-                            fs.remove_node(Node::MODE_SYMLINK, child_name, parent.0)
+                    if !child.data().is_dir() {
+                        if child.data().uid() != uid {
+                            // println!("file not owned by current user {}", parent.1.uid);
+                            return Err(Error::new(EACCES));
+                        }
+
+                        if child.data().is_symlink() {
+                            tx.remove_node(parent.ptr(), &child_name, Node::MODE_SYMLINK)
                                 .and(Ok(0))
                         } else {
-                            fs.remove_node(Node::MODE_FILE, child_name, parent.0)
+                            tx.remove_node(parent.ptr(), &child_name, Node::MODE_FILE)
                                 .and(Ok(0))
                         }
                     } else {
-                        Err(Error::new(ENOENT))
+                        Err(Error::new(EISDIR))
                     }
                 } else {
-                    Err(Error::new(EISDIR))
+                    Err(Error::new(EPERM))
                 }
             } else {
-                Err(Error::new(EPERM))
+                Err(Error::new(ENOENT))
             }
-        } else {
-            Err(Error::new(ENOENT))
-        }
+        })
     }
 
     /* Resource operations */
     #[allow(unused_variables)]
-    fn dup(&self, old_id: usize, buf: &[u8]) -> Result<usize> {
+    fn dup(&mut self, old_id: usize, buf: &[u8]) -> Result<usize> {
         // println!("Dup {}", old_id);
 
         if !buf.is_empty() {
             return Err(Error::new(EINVAL));
         }
 
-        let mut files = self.files.borrow_mut();
-        let resource = if let Some(old_resource) = files.get(&old_id) {
+        let resource = if let Some(old_resource) = self.files.get(&old_id) {
             old_resource.dup()?
         } else {
             return Err(Error::new(EBADF));
         };
 
         let id = self.next_id.fetch_add(1, Ordering::SeqCst);
-        files.insert(id, resource);
+        self.files.insert(id, resource);
 
         Ok(id)
     }
 
     #[allow(unused_variables)]
-    fn read(&self, id: usize, buf: &mut [u8]) -> Result<usize> {
+    fn read(&mut self, id: usize, buf: &mut [u8]) -> Result<usize> {
         // println!("Read {}, {:X} {}", id, buf.as_ptr() as usize, buf.len());
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.read(buf, &mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.read(buf, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn write(&self, id: usize, buf: &[u8]) -> Result<usize> {
+    fn write(&mut self, id: usize, buf: &[u8]) -> Result<usize> {
         // println!("Write {}, {:X} {}", id, buf.as_ptr() as usize, buf.len());
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.write(buf, &mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.write(buf, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn seek(&self, id: usize, pos: isize, whence: usize) -> Result<isize> {
+    fn seek(&mut self, id: usize, pos: isize, whence: usize) -> Result<isize> {
         // println!("Seek {}, {} {}", id, pos, whence);
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.seek(pos, whence, &mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.seek(pos, whence, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn fchmod(&self, id: usize, mode: u16) -> Result<usize> {
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.fchmod(mode, &mut self.fs.borrow_mut())
+    fn fchmod(&mut self, id: usize, mode: u16) -> Result<usize> {
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.fchmod(mode, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn fchown(&self, id: usize, uid: u32, gid: u32) -> Result<usize> {
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.fchown(uid, gid, &mut self.fs.borrow_mut())
+    fn fchown(&mut self, id: usize, uid: u32, gid: u32) -> Result<usize> {
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.fchown(uid, gid, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn fcntl(&self, id: usize, cmd: usize, arg: usize) -> Result<usize> {
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
+    fn fcntl(&mut self, id: usize, cmd: usize, arg: usize) -> Result<usize> {
+        if let Some(file) = self.files.get_mut(&id) {
             file.fcntl(cmd, arg)
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn fevent(&self, id: usize, _flags: EventFlags) -> Result<EventFlags> {
-        let files = self.files.borrow_mut();
-        if let Some(file) = files.get(&id) {
+    fn fevent(&mut self, id: usize, _flags: EventFlags) -> Result<EventFlags> {
+        if let Some(_file) = self.files.get(&id) {
             // EPERM is returned for files that are always readable or writable
             Err(Error::new(EPERM))
         } else {
@@ -547,10 +571,9 @@ impl<D: Disk> Scheme for FileScheme<D> {
         }
     }
 
-    fn fpath(&self, id: usize, buf: &mut [u8]) -> Result<usize> {
+    fn fpath(&mut self, id: usize, buf: &mut [u8]) -> Result<usize> {
         // println!("Fpath {}, {:X} {}", id, buf.as_ptr() as usize, buf.len());
-        let files = self.files.borrow_mut();
-        if let Some(file) = files.get(&id) {
+        if let Some(file) = self.files.get(&id) {
             let name = self.name.as_bytes();
 
             let mut i = 0;
@@ -567,13 +590,23 @@ impl<D: Disk> Scheme for FileScheme<D> {
                 i += 1;
             }
 
-            file.path(&mut buf[i..]).map(|count| i + count)
+            let path = file.path().as_bytes();
+            let mut j = 0;
+            while i < buf.len() && j < path.len() {
+                buf[i] = path[j];
+                i += 1;
+                j += 1;
+            }
+
+            Ok(i)
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn frename(&self, id: usize, url: &str, uid: u32, gid: u32) -> Result<usize> {
+    fn frename(&mut self, id: usize, url: &str, uid: u32, gid: u32) -> Result<usize> {
+        unimplemented!();
+        /*TODO: FRENAME
         let path = url.trim_matches('/');
 
         // println!("Frename {}, {} from {}, {}", id, path, uid, gid);
@@ -613,19 +646,19 @@ impl<D: Disk> Scheme for FileScheme<D> {
                 }
 
                 if let Some(ref node) = node_opt {
-                    if !node.1.owner(uid) {
+                    if !node.data().owner(uid) {
                         // println!("new dir not owned by caller {}", uid);
                         return Err(Error::new(EACCES));
                     }
 
-                    if node.1.is_dir() {
+                    if node.data().is_dir() {
                         if !orig.1.is_dir() {
                             // println!("orig is file, new is dir");
                             return Err(Error::new(EACCES));
                         }
 
                         let mut children = Vec::new();
-                        fs.child_nodes(&mut children, node.0)?;
+                        fs.child_nodes(node.ptr(), &mut children)?;
 
                         if !children.is_empty() {
                             // println!("new dir not empty");
@@ -670,29 +703,23 @@ impl<D: Disk> Scheme for FileScheme<D> {
         } else {
             Err(Error::new(EBADF))
         }
+        */
     }
 
-    fn fstat(&self, id: usize, stat: &mut Stat) -> Result<usize> {
+    fn fstat(&mut self, id: usize, stat: &mut Stat) -> Result<usize> {
         // println!("Fstat {}, {:X}", id, stat as *mut Stat as usize);
-        let files = self.files.borrow_mut();
-        if let Some(file) = files.get(&id) {
-            file.stat(stat, &mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get(&id) {
+            self.fs.tx(|tx| file.stat(stat, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn fstatvfs(&self, id: usize, stat: &mut StatVfs) -> Result<usize> {
-        let files = self.files.borrow_mut();
-        if let Some(_file) = files.get(&id) {
-            let mut fs = self.fs.borrow_mut();
-
-            let free = fs.header.1.free;
-            let free_size = fs.node_len(free)?;
-
+    fn fstatvfs(&mut self, id: usize, stat: &mut StatVfs) -> Result<usize> {
+        if let Some(_file) = self.files.get(&id) {
             stat.f_bsize = BLOCK_SIZE as u32;
-            stat.f_blocks = fs.header.1.size / (stat.f_bsize as u64);
-            stat.f_bfree = free_size / (stat.f_bsize as u64);
+            stat.f_blocks = self.fs.header.size() / (stat.f_bsize as u64);
+            stat.f_bfree = self.fs.allocator().free();
             stat.f_bavail = stat.f_bfree;
 
             Ok(0)
@@ -701,53 +728,48 @@ impl<D: Disk> Scheme for FileScheme<D> {
         }
     }
 
-    fn fsync(&self, id: usize) -> Result<usize> {
+    fn fsync(&mut self, id: usize) -> Result<usize> {
         // println!("Fsync {}", id);
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.sync(&mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.sync(tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn ftruncate(&self, id: usize, len: usize) -> Result<usize> {
+    fn ftruncate(&mut self, id: usize, len: usize) -> Result<usize> {
         // println!("Ftruncate {}, {}", id, len);
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.truncate(len, &mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.truncate(len, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn futimens(&self, id: usize, times: &[TimeSpec]) -> Result<usize> {
+    fn futimens(&mut self, id: usize, times: &[TimeSpec]) -> Result<usize> {
         // println!("Futimens {}, {}", id, times.len());
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            file.utimens(times, &mut self.fs.borrow_mut())
+        if let Some(file) = self.files.get_mut(&id) {
+            self.fs.tx(|tx| file.utimens(times, tx))
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn fmap(&self, id: usize, map: &Map) -> Result<usize> {
+    fn fmap(&mut self, id: usize, map: &Map) -> Result<usize> {
         // println!("Fmap {}, {:?}", id, map);
-        let mut files = self.files.borrow_mut();
-        if let Some(file) = files.get_mut(&id) {
-            let address = file.fmap(map, &mut self.fs.borrow_mut())?;
-            self.fmap.borrow_mut().insert(address, id);
+        if let Some(file) = self.files.get_mut(&id) {
+            let address = self.fs.tx(|tx| file.fmap(map, tx))?;
+            self.fmap.insert(address, id);
             Ok(address)
         } else {
             Err(Error::new(EBADF))
         }
     }
 
-    fn funmap_old(&self, address: usize) -> Result<usize> {
-        if let Some(id) = self.fmap.borrow_mut().remove(&address) {
-            let mut files = self.files.borrow_mut();
-            if let Some(file) = files.get_mut(&id) {
-                file.funmap(address, &mut self.fs.borrow_mut())
+    fn funmap_old(&mut self, address: usize) -> Result<usize> {
+        if let Some(id) = self.fmap.remove(&address) {
+            if let Some(file) = self.files.get_mut(&id) {
+                self.fs.tx(|tx| file.funmap(address, tx))
             } else {
                 Err(Error::new(EINVAL))
             }
@@ -756,16 +778,23 @@ impl<D: Disk> Scheme for FileScheme<D> {
         }
     }
 
-    //TODO: implement
-    fn funmap(&self, address: usize, length: usize) -> Result<usize> {
+    //TODO: implement (length is ignored!)
+    fn funmap(&mut self, address: usize, length: usize) -> Result<usize> {
         println!("redoxfs: funmap 0x{:X}, {}", address, length);
-        Err(Error::new(ENOSYS))
+        if let Some(id) = self.fmap.remove(&address) {
+            if let Some(file) = self.files.get_mut(&id) {
+                self.fs.tx(|tx| file.funmap(address, tx))
+            } else {
+                Err(Error::new(EINVAL))
+            }
+        } else {
+            Err(Error::new(EINVAL))
+        }
     }
 
-    fn close(&self, id: usize) -> Result<usize> {
+    fn close(&mut self, id: usize) -> Result<usize> {
         // println!("Close {}", id);
-        let mut files = self.files.borrow_mut();
-        if files.remove(&id).is_some() {
+        if self.files.remove(&id).is_some() {
             Ok(0)
         } else {
             Err(Error::new(EBADF))
